import unittest

import numpy as np
import numpy.testing as npt
from openmdao.api import Group, Problem

import wisdem.plant_financese.plant_finance as pf


class TestPlantFinance(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.discrete_inputs = {}
        self.outputs = {}

        self.inputs["machine_rating"] = 1000.0
        self.inputs["tcc_per_kW"] = 1.2e3
        self.inputs["offset_tcc_per_kW"] = 0.0
        self.inputs["bos_per_kW"] = 7.7e3
        self.inputs["opex_per_kW"] = 7e2
        self.inputs["fixed_charge_rate"] = 0.12
        self.inputs["turbine_aep"] = 1.6e7
        self.inputs["plant_aep_in"] = 1.6e7 * 50
        self.inputs["wake_loss_factor"] = 0.15

        self.discrete_inputs["turbine_number"] = 50

        self.inputs["electricity_price"] = 0.04
        self.inputs["reserve_margin_price"] = 120
        self.inputs["capacity_credit"] = 1.0
        self.inputs["benchmark_price"] = 0.071

        self.mypfin = pf.PlantFinance(verbosity=True)

    def testRun(self):
        # Plant AEP way
        self.mypfin.compute(self.inputs, self.outputs, self.discrete_inputs, {})

        C = 0.12 * (1.2e3 + 7.7e3) + 7e2
        E = (1.6e7 * 50.0) / (1e3 * 50)
        V = (1.6e7 * 50.0) / (1e3 * 50) * 0.04 + 120.0 * 1.0
        lcoe = C / E
        plcoe = C / V * 0.071
        self.assertEqual(self.outputs["lcoe"], lcoe)
        self.assertEqual(self.outputs["plcoe"], plcoe)

        # Wake loss way
        self.inputs["plant_aep_in"] = 0.0
        self.mypfin.compute(self.inputs, self.outputs, self.discrete_inputs, {})

        C = 0.12 * (1.2e3 + 7.7e3) + 7e2
        E = (1.6e7 * 50.0 * (1 - 0.15)) / (1e3 * 50)
        V = (1.6e7 * 50.0 * (1 - 0.15)) / (1e3 * 50) * 0.04 + 120.0 * 1.0
        lcoe = C / E
        plcoe = C / V * 0.071
        self.assertEqual(self.outputs["lcoe"], lcoe)
        self.assertEqual(self.outputs["plcoe"], plcoe)

    def testDerivatives(self):
        prob = Problem(reports=False)
        root = prob.model = Group()
        root.add_subsystem("pf", pf.PlantFinance(), promotes=["*"])
        prob.setup()
        for k in self.inputs.keys():
            prob[k] = self.inputs[k]
        for k in self.discrete_inputs.keys():
            prob[k] = self.discrete_inputs[k]
        # prob.check_partials()


if __name__ == "__main__":
    unittest.main()



import matplotlib.pyplot as plt

from wisdem.glue_code.runWISDEM import load_wisdem

refturb, _, _ = load_wisdem("outputs/refturb_output.pkl")
xs = refturb["blade.outer_shape_bem.s_default"]
ys = refturb["rotorse.rp.powercurve.ax_induct_regII"]
fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 5))
ax.plot(xs, ys)
ax.set_xlabel("Blade Nondimensional Span [-]")
ax.set_ylabel("Axial Induction [-]")
plt.show()



import pandas as pd

class XlsxValidator:
    """
    XlsxValidator is for comparing the results of a previous model run
    to the results of a current model run.
    """

    def compare_expected_to_actual(self, expected_xlsx, actual_module_type_operation_list, validation_output_xlsx):
        """
        This compares the expected costs as calculated by a prior model run
        with the actual results from a current model run.

        It compares the results row by row and prints any differences.

        Parameters
        ----------
        expected_xlsx : str
            The absolute filename of the expected output .xlsx file.

        actual_module_type_operation_list : str
            The module_type_operation_list as returned by a subclass of
            XlsxManagerRunner.

        validation_output_xlsx : str
            The absolute pathname to the output file with the comparison
            results.

        Returns
        -------
        bool
            True if the expected and actual results are equal. It returns
            False otherwise.
        """
        # First, make the list of dictionaries into a dataframe, and drop
        # the raw_cost and raw_cost_total_or_per_turbine columns.
        actual_df = pd.DataFrame(actual_module_type_operation_list)
        actual_df.drop(['raw_cost', 'raw_cost_total_or_per_turbine'], axis=1, inplace=True)
        expected_df = pd.read_excel(expected_xlsx, 'costs_by_module_type_operation', engine='openpyxl')
        #expected_df = expected_df.dropna(inplace=True, how='all')
        expected_df.rename(columns={
            'Project ID with serial': 'project_id_with_serial',
            'Number of turbines': 'num_turbines',
            'Turbine rating MW': 'turbine_rating_MW',
            'Module': 'module',
            'Operation ID': 'operation_id',
            'Type of cost': 'type_of_cost',
            'Cost per turbine': 'cost_per_turbine',
            'Cost per project': 'cost_per_project',
            'USD/kW per project': 'usd_per_kw_per_project'
        }, inplace=True)

        cost_per_project_actual = actual_df[
            ['cost_per_project', 'project_id_with_serial', 'module', 'operation_id', 'type_of_cost']]
        cost_per_project_expected = expected_df[
            ['cost_per_project', 'project_id_with_serial', 'module', 'operation_id', 'type_of_cost']]

        comparison = cost_per_project_actual.merge(
            cost_per_project_expected,
            on=['project_id_with_serial', 'module', 'operation_id', 'type_of_cost'])

        comparison.rename(columns={'cost_per_project_x': 'cost_per_project_actual',
                                    'cost_per_project_y': 'cost_per_project_expected'}, inplace=True)

        comparison['difference_validation'] = comparison['cost_per_project_actual'] - comparison['cost_per_project_expected']

        # Regardless of the outcome, write the end result of the comparison
        # to the validation output file.
        columns_for_comparison_output = [
            'project_id_with_serial',
            'module',
            'operation_id',
            'type_of_cost',
            'cost_per_project_actual',
            'cost_per_project_expected',
            'difference_validation'
        ]
        comparison.to_excel(validation_output_xlsx, index=False, columns=columns_for_comparison_output)

        # If the comparison dataframe is empty, that means there are no common
        # projects in the expected data that match the actual data.
        if len(comparison) < 1:
            print('=' * 80)
            print('Validation error: There are no common projects between actual and expected data.')
            print('=' * 80)
            return False

        # Find all rows where the difference is unequal to 0. These are rows
        # that failed validation. Note that, after the join, the rows may be
        # in a different order than the originals.
        #
        # Round the difference to a given number of decimal places.
        failed_rows = comparison[comparison['difference_validation'].round(decimals=4) != 0]

        if len(failed_rows) > 0:
            print('=' * 80)
            print('The following rows failed validation:')
            print(failed_rows)
            print('=' * 80)
            return False
        else:
            return True



import os
import re
from typing import Any, Dict, List, Union
import pandas as pd
import numpy as np

try:
    import dearpygui.dearpygui as dpg
    GUI_AVAIL = True
except ImportError:
    GUI_AVAIL = False

import wisdem.inputs.validation as val
from wisdem.glue_code.runWISDEM import run_wisdem


def _hsv_to_rgb(h, s, v):
    if s == 0.0:
        return (v, v, v)
    i = int(h * 6.0)  # XXX assume int() truncates!
    f = (h * 6.0) - i
    p, q, t = v * (1.0 - s), v * (1.0 - s * f), v * (1.0 - s * (1.0 - f))
    i %= 6
    if i == 0:
        return (255 * v, 255 * t, 255 * p)
    if i == 1:
        return (255 * q, 255 * v, 255 * p)
    if i == 2:
        return (255 * p, 255 * v, 255 * t)
    if i == 3:
        return (255 * p, 255 * q, 255 * v)
    if i == 4:
        return (255 * t, 255 * p, 255 * v)
    if i == 5:
        return (255 * v, 255 * p, 255 * q)
    

class DPGLineEdit(object):
    """
    Introduces a value edit on a line for the yaml dictionary leafs.  Updates referenced dictionary on callback.

    """

    def __init__(self, dictionary: Union[List[Any], Dict[str, Any]], key: str, value: Any, parent_tag: str) -> None:
        """
        The instance attributes for this class are:

        _mydict: Dict[str, Any]
            The dictionary to be changed when this text field is changed.

        _mykey: str
            The key on the dictionary to be changed when this text field
            changes.

        _list_re: re.Pattern
            The regex that matches a string that contains a list, so that
            it does not need to be instantiated each time it is used.

        Parameters
        ----------
        dictionary: Dict[str, Any]
            The dictionary to be changed.

        key: str
            The key whose value is to be changed on the dictionary.
        """
        self._mydict = dictionary
        self._mykey = key
        self._list_re = re.compile(r"^\[.*\]$")

        if self.is_list(str(value)):
            dpg.add_input_text(label=key, default_value=str(list(value)), callback=self.mycallback, parent=parent_tag)

        elif self.is_float(value):
            dpg.add_input_float(
                label=key, default_value=float(value), width=200, callback=self.mycallback, parent=parent_tag
            )

        elif self.is_boolean(value):
            dpg.add_input_text(
                label=key, default_value=str(bool(value)), width=150, callback=self.mycallback, parent=parent_tag
            )

        elif self.is_integer(value):
            # (have to list after is_boolean otherwise T/F becomes 1/0)
            dpg.add_input_int(
                label=key, default_value=int(value), width=200, callback=self.mycallback, parent=parent_tag
            )

        else:
            dpg.add_input_text(
                label=key, default_value=str(value), width=300, callback=self.mycallback, parent=parent_tag
            )

    def mycallback(self, sender, app_data) -> None:
        """
        Callback on line item entry to update referenced dictionary
        """
        if self._mydict is not None and self._mykey is not None:
            if self.is_list(str(app_data)):
                value = self.parse_list(app_data)

            elif self.is_float(app_data):
                value = float(app_data)

            elif self.is_integer(app_data):
                value = int(app_data)

            elif self.is_boolean(app_data):
                value = self.parse_boolean(app_data)

            else:
                value = str(app_data)

            self._mydict[self._mykey] = value

    def parse_list(self) -> List[Union[str, float]]:
        """
        This parses the text in the field to a list of numbers and/or strings.

        Returns
        -------
        List[Union[str, float]]
            A list that contains strings and floats, depending on what could be
            parsed out of the text in the field.
        """
        trimmed_text = self.text()[1:-1]
        trimmed_str_values = [x.strip() for x in trimmed_text.split(",")]
        result: List[Union[str, float]] = []
        for x in trimmed_str_values:
            if self.is_integer(x):
                result.append(int(x))
            elif self.is_float(x):
                result.append(float(x))
            else:
                result.append(x)
        return result

    @staticmethod
    def parse_boolean(value: str) -> bool:
        """
        This method parses a string as a boolean. See the is_boolean()
        method below.

        Parameters
        ----------
        value: str
            The value that is being parsed as a boolean.

        Returns
        -------
        bool
            The value, parsed as a boolean.
        """

        return True if str(value).lower() == "true" else False

    @staticmethod
    def is_float(value: Any) -> bool:
        """
        This tests if a value is a float and returns True or False depending
        on the outcome of the test

        Parameters
        ----------
        value: Any
            The value to test.

        Returns
        -------
        bool
            True if the value is a float, False otherwise.
        """
        try:
            float(value)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_integer(value: Any) -> bool:
        """
        This tests if a value is an integer and returns True or False depending
        on the outcome of the test

        Parameters
        ----------
        value: Any
            The value to test.

        Returns
        -------
        bool
            True if the value is an integer, False otherwise.
        """
        try:
            int(value)
            return True
        except ValueError:
            return False

    def is_list(self, value: str) -> bool:
        """
        Determines whether a value is a list through a regular expression
        match. The intent is that, if the value is a list, it will be
        parsed as a list.

        Parameters
        ----------
        value: str

        Returns
        -------
        bool
            True if the value appears to be a list, false otherwise
        """
        return self._list_re.match(value) is not None

    def is_boolean(self, value: str) -> bool:
        """
        Determines whether a string encodes a "True" or a "False" value. It
        does not parse the boolean value. If this method returns True, then
        the value can be parsed as a True or False boolean.

        Both Pythonic True and False and JSON-style true and false are
        accepted as booleans.

        Note: This should probably be checkbox, not a text box, but this
        is a temporary workaround.

        Parameters
        ----------
        value: str
            The string to be tested.

        Returns
        -------
        bool
            True if the string represents a boolean, false otherwise.
        """
        return str(value).lower() in ["true", "false"]


class GUI_Master(object):
    def __init__(self):
        self._input_mode = True

        self.working_dir = os.path.expanduser("~")
        self.froot_export = "gui_export"

        self.geometry_dict = {}
        self.modeling_dict = {}
        self.analysis_dict = {}

        self.output_data = None
        
    def _on_demo_close(self):
        pass

    def _get_file_names(self):
        fname_geom = f"{self.froot_export}_geometry.yaml"
        fname_model = f"{self.froot_export}_modeling.yaml"
        fname_anal = f"{self.froot_export}_analysis.yaml"
        return fname_geom, fname_model, fname_anal

    def _write_files(self):
        """
        The "Run WEIS" click event handler calls this method when it is ready
        to make an attempt to run WEIS. It checks to see if all file shave been
        edited
        """
        fname_geom, fname_model, fname_anal = self._get_file_names()
        val.write_yaml(self.geometry_dict, fname_geom)
        val.write_yaml(self.modeling_dict, fname_model)
        val.write_yaml(self.analysis_dict, fname_anal)

    def _recursion_dict_display(self, dict_or_list, parent_tag=""):
        """
        This recursive method is where the automatic layout magic happens.
        This method calls itself recursively as it descends down the dictionary
        nesting structure.

        Basically, any given dictionary can consist of scalar and dictionary
        values. At each level of the dictionary, edit fields are placed for
        scalar values and tabbed widgets are placed for the next level of
        nesting.

        Parameters
        ----------
        dict_or_list: Dict[str, Any]
            The dictionary to automatically lay out in to the interface.
        """
        subscripts_values = dict_or_list.items() if type(dict_or_list) is dict else enumerate(dict_or_list)  # type: ignore

        for k, v in subscripts_values:
            # Create the unique tag for this item
            itag = dpg.generate_uuid()

            # Recursive call for nested dictionaries within dictionaries.
            if type(v) is dict:
                dpg.add_tree_node(label=k, tag=itag, parent=parent_tag)
                self._recursion_dict_display(v, parent_tag=itag)

            # Recursive call for nested dictionaries within lists.
            elif type(v) is list and len(v) > 0 and type(v[0]) is dict:
                dpg.add_tree_node(label=k, tag=itag, parent=parent_tag, bullet=True)
                self._recursion_dict_display(v, parent_tag=itag)

            # Otherwise just lay out a label and text field.
            else:
                DPGLineEdit(dict_or_list, k, v, parent_tag)

    def _run_program(self):
        with dpg.window(modal=True, no_close=True, popup=True, show=True, autosize=True, tag="run_popup"):
            dpg.add_text("Running WISDEM in terminal or console window")

            current_dir = os.getcwd()
            os.chdir(self.working_dir)

            fname_geom, fname_model, fname_anal = self._get_file_names()
            self._write_files()
            run_wisdem(fname_geom, fname_model, fname_anal)

            os.chdir(current_dir)

        dpg.configure_item("run_popup", show=False)

    def _export_root(self, sender, app_data):
        self.froot_export = app_data

    def _read_output(self, sender, app_data):
        if app_data is not None:
            # Grab the filename selected
            temp_dict = app_data["selections"]
            fname = [temp_dict[m] for m in temp_dict][0]

            # Read in the file
            if fname.lower().endswith('.pkl'):
                temp_data = pd.read_pickle(fname)
            elif fname.lower().endswith('.csv'):
                temp_data = pd.read_csv(fname)
            elif fname.lower().endswith('.xlsx'):
                temp_data = pd.read_excel(fname)

            # Fancy way of dropping openmdao duplicates
            self.output_data = temp_data.loc[temp_data.astype(str).drop_duplicates(subset=['units','values']).index]

            # Loop over all data and cleanup
            temp_size = np.zeros(len(self.output_data), dtype=np.int_)
            temp_val = [0]*len(self.output_data)
            var_full  = ['']*len(self.output_data)
            to_delete = []
            for k in range(len(self.output_data)):
                ivar = self.output_data['variables'].iloc[k]
                ival = self.output_data['values'].iloc[k]
                iunit = self.output_data['units'].iloc[k]

                # csv and xlsx bring in values as string- convert to python types
                if isinstance(ival, str):
                    try:
                        temp_val[k] = eval(ival)
                        ival = temp_val[k]
                    except:
                        to_delete.append(self.output_data.index[k])
                else:
                    temp_val[k] = ival

                # Cleanup no-unit entries
                if str(iunit).lower() in ['','na','n/a','none','nan']:
                    iunit = '-'

                # Identify size
                try:
                    temp_size[k] = len(ival)
                except:
                    temp_size[k] = 1

                # Variable axis/listbox name
                var_full[k] = f'{ivar} [{iunit}] ({temp_size[k]})'

            # Store clean data
            self.output_data['size'] = temp_size
            self.output_data['variables_full'] = var_full
            self.output_data['values'] = temp_val

            # Drop bad data, then sort
            self.output_data.drop(index=to_delete, inplace=True)
            self.output_data.sort_values('variables_full', inplace=True)

            # Update list box
            dpg.configure_item("xylistlabel", show=True)
            dpg.configure_item("xylist", show=True)
            dpg.configure_item("plotbutton", show=True)
            dpg.configure_item("xlist", items=self.output_data['variables_full'].to_list(), show=True)
            dpg.configure_item("ylist", items=self.output_data['variables_full'].to_list(), show=True)
            
    def _load_output(self):
        with dpg.file_dialog(label="WISDEM Output File Selector", directory_selector=False, show=True,
                             callback=self._read_output, width=500, height=400, default_path=self.working_dir):
            dpg.add_file_extension(".*")
            dpg.add_file_extension(".xlsx", color=(0, 255, 0, 255), custom_text="[Excel]")
            dpg.add_file_extension(".csv", color=(0, 255, 0, 255), custom_text="[CSV]")
            dpg.add_file_extension(".pkl", color=(0, 255, 0, 255), custom_text="[Pickle]")

    def _set_workdir(self, sender, app_data):
        if app_data is not None:
            self.working_dir = app_data["current_path"].strip()
        dpg.configure_item("workdir_field", default_value=f"Working directory: {self.working_dir}")

    def _choose_workdir(self):
        dpg.add_file_dialog(
            label="Working Directory Selector",
            directory_selector=True,
            show=True,
            width=500,
            height=400,
            callback=self._set_workdir,
            default_path=self.working_dir,
        )

    def _set_file_field(self, sender, app_data, user_data):
        if not app_data is None:
            fpath = app_data["file_path_name"].strip()

            if user_data.lower().find("geometry") >= 0:
                self.geometry_dict = val.load_geometry_yaml(fpath)
                mydict = self.geometry_dict
                id_root = "geometry"

            elif user_data.lower().find("modeling") >= 0:
                self.modeling_dict = val.load_modeling_yaml(fpath)
                mydict = self.modeling_dict
                id_root = "modeling"

            elif user_data.lower().find("analysis") >= 0:
                self.analysis_dict = val.load_analysis_yaml(fpath)
                mydict = self.analysis_dict
                id_root = "analysis"

        # Delete the old nested yaml file and add the new one
        obj_id = id_root + "_context"  # user_data.replace("field", "context")
        dpg.delete_item(obj_id, children_only=True)
        self._recursion_dict_display(mydict, parent_tag=obj_id)

    def _choose_file(self, sender, app_data, user_data):
        with dpg.file_dialog(
            label="YAML File Selector",
            directory_selector=False,
            show=True,
            width=500,
            height=400,
            user_data=user_data,
            callback=self._set_file_field,
        ):
            dpg.add_file_extension("YAML Files (*.yml *.yaml){.yml,.yaml}", color=(0, 255, 255, 255))
            dpg.add_file_extension(".*", color=(255, 255, 255, 255))

    def _variable_list(self, sender, app_data, user_data):
        # When enter x-value in listbox, reduce y-values to equivalent size kosher for plotting
        xval_full = dpg.get_value("xselected")
        xdf = self.output_data[ self.output_data['variables_full'] == xval_full ]
        xsize = int(xdf['size'].iloc[0])
        yreduced = self.output_data[ self.output_data['size']==xsize ]
        dpg.configure_item("ylist", items=yreduced['variables_full'].to_list(), show=True)

    def _gen_plot(self, sender, app_data, user_data):
        # Build the plot, reset the axis
        xval_full = dpg.get_value("xselected")
        yval_full = dpg.get_value("yselected")
        xstr = " ".join(xval_full.split()[:-1])
        ystr = " ".join(yval_full.split()[:-1])
        xdf = self.output_data[ self.output_data['variables_full'] == xval_full ]
        ydf = self.output_data[ self.output_data['variables_full'] == yval_full ]
        xx = np.array(xdf['values'].to_list()).flatten()
        yy = np.array(ydf['values'].to_list()).flatten()
        dpg.configure_item("xaxis", label=xstr)
        dpg.configure_item("yaxis", label=ystr)
        dpg.configure_item("lineseries", x=xx, y=yy)
        dpg.reset_axis_zoom_constraints('xaxis')
        dpg.reset_axis_zoom_constraints('yaxis')
        dpg.reset_axis_limits_constraints('xaxis')
        dpg.reset_axis_limits_constraints('yaxis')
        dpg.fit_axis_data('xaxis')
        dpg.fit_axis_data('yaxis')
        dpg.reset_axis_ticks('xaxis')
        dpg.reset_axis_ticks('yaxis')
        dpg.configure_item("myplot", show=True)
                      
    
    def show_gui(self):
        def _log(sender, app_data, user_data):
            print(f"sender: {sender}, \t app_data: {app_data}, \t user_data: {user_data}")

        # GUI variables for selectors
        with dpg.value_registry():
            dpg.add_string_value(default_value="x-axis string", tag="xselected")
            dpg.add_string_value(default_value="y-axis string", tag="yselected")
            
        # Make buttons stand out a big more
        with dpg.theme(tag="button_theme"):
            with dpg.theme_component(dpg.mvButton):
                dpg.add_theme_color(dpg.mvThemeCol_Button, _hsv_to_rgb(3.0 / 7.0, 0.6, 0.6))
                dpg.add_theme_color(dpg.mvThemeCol_ButtonActive, _hsv_to_rgb(3.0 / 7.0, 0.8, 0.8))
                dpg.add_theme_color(dpg.mvThemeCol_ButtonHovered, _hsv_to_rgb(3.0 / 7.0, 0.7, 0.7))
                dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, 5)
                dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 3, 3)

        # INPUT WINDOW
        with dpg.window(width=750, height=800, on_close=self._on_demo_close, pos=(0, 0), label='Input'):
            with dpg.menu_bar():
                #with dpg.menu(label="Mode"):
                #    dpg.add_combo( ("Input Editor", "Output Viewer"), width=140, default_value="Input Editor",
                #                   label="Mode", tag="mode_dropdown", callback=self._mode_set)

                with dpg.menu(label="Files"):
                    dpg.add_menu_item(label="Set Working Diretory...", callback=self._choose_workdir)

                    with dpg.menu(label="Import"):
                        dpg.add_menu_item(label="Geometry YAML", user_data="geometry", callback=self._choose_file)
                        dpg.add_menu_item(label="Modeling YAML", user_data="modeling", callback=self._choose_file)
                        dpg.add_menu_item(label="Analysis YAML", user_data="analysis", callback=self._choose_file)
                        # dpg.add_menu_item(label="Output NPZ")

                    dpg.add_menu_item(label="Export Input Files...")

                with dpg.menu(label="Run"):
                    dpg.add_menu_item(label="Run WISDEM/WEIS...", callback=self._run_program)

            # Display and/or change working directory
            dpg.add_text("")
            
            with dpg.group(horizontal=True):
                dpg.add_text("Temp", tag="workdir_field")
                dpg.add_button(label="Change", callback=self._choose_workdir)
                dpg.bind_item_theme(dpg.last_item(), "button_theme")

            with dpg.group(horizontal=True):
                dpg.add_text("Export Prefix:")
                dpg.add_input_text(label="", default_value=self.froot_export, width=200, callback=self._export_root, tag="export_field")
                
            dpg.add_text("")
            dpg.add_text("WISDEM/WEIS YAML-Based Input Files:", color=(255, 255, 0))

            with dpg.tab_bar():
                with dpg.tab(label="Geometry"):
                    dpg.add_button(label="Import YAML File", user_data="geometry", callback=self._choose_file)
                    dpg.bind_item_theme(dpg.last_item(), "button_theme")
                    dpg.add_text("")
                    dpg.add_collapsing_header(label="WindIO Geometry", tag="geometry_context")

                with dpg.tab(label="Modeling"):
                    dpg.add_button(label="Import YAML File", user_data="modeling", callback=self._choose_file)
                    dpg.bind_item_theme(dpg.last_item(), "button_theme")
                    dpg.add_text("")
                    dpg.add_collapsing_header(label="WISDEM/WEIS Modeling", tag="modeling_context")

                with dpg.tab(label="Analysis"):
                    dpg.add_button(label="Import YAML File", user_data="analysis", callback=self._choose_file)
                    dpg.bind_item_theme(dpg.last_item(), "button_theme")
                    dpg.add_text("")
                    dpg.add_collapsing_header(label="WISDEM/WEIS Analysis", tag="analysis_context")

        # Do init steps
        temp_dir = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))),
            "examples",
            "02_reference_turbines",
        )
        temp_geom = os.path.join(temp_dir, "IEA-15-240-RWT.yaml")
        temp_model = os.path.join(temp_dir, "modeling_options.yaml")
        temp_anal = os.path.join(temp_dir, "analysis_options.yaml")

        self._set_workdir(None, None)
        self._set_file_field(None, {"file_path_name": temp_geom}, "geometry_field")
        self._set_file_field(None, {"file_path_name": temp_model}, "modeling_field")
        self._set_file_field(None, {"file_path_name": temp_anal}, "analysis_field")

        # OUTPUT WINDOW
        with dpg.window(width=900, height=800, on_close=self._on_demo_close, pos=(750, 0), label='Output'):
            
            with dpg.menu_bar():
                with dpg.menu(label="Load"):
                    dpg.add_menu_item(label="Select output file", callback=self._load_output)

            with dpg.group(horizontal=True, tag='xylistlabel', show=False):
                dpg.add_text("X-Axis Variable")
                dpg.add_text(" "*50)
                dpg.add_text("Y-Axis Variable")
                
            with dpg.group(horizontal=True, tag='xylist', show=False):
                dpg.add_listbox([], label='', tag='xlist', width=450, source="xselected", num_items=15, callback=self._variable_list)
                dpg.add_listbox([], label='', tag='ylist', width=450, source="yselected", num_items=15)

            with dpg.group(horizontal=True, tag='plotbutton', show=False):
                dpg.add_button(label="Line Plot", callback=self._gen_plot, user_data="line")
                dpg.bind_item_theme(dpg.last_item(), "button_theme")
                #dpg.add_button(label="Bar Plot", callback=self._gen_plot, user_data="bar")
                #dpg.bind_item_theme(dpg.last_item(), "button_theme")

            with dpg.plot(label="WISDEM Plot", height=400, width=-1, show=False, tag='myplot'):
                #dpg.add_plot_legend()
                dpg.add_plot_axis(dpg.mvXAxis, label="x", tag='xaxis')
                with dpg.plot_axis(dpg.mvYAxis, label="y", tag='yaxis'):
                    # GUI crashes if these are numpy arrays, but lists seem to work.  Don't know why.
                    dpg.add_line_series(np.arange(10).tolist(), np.arange(10).tolist(), parent='yaxis', tag='lineseries')


def run():
    if GUI_AVAIL:
        dpg.create_context()
        dpg.create_viewport(title="NREL's WISDEM/WEIS Input/Output GUI", width=1650, height=800)

        mygui = GUI_Master()
        mygui.show_gui()

        dpg.setup_dearpygui()
        dpg.show_viewport()
        dpg.start_dearpygui()
        dpg.destroy_context()
    else:
        print("GUI is not available")

if __name__ == "__main__":
    run()



import copy

import numpy as np
from openmdao.api import Group, ExplicitComponent
from scipy.interpolate import PchipInterpolator

from wisdem.precomp import PreComp, Profile, CompositeSection, Orthotropic2DMaterial
from wisdem.commonse.utilities import arc_length
from wisdem.precomp.precomp_to_beamdyn import pc2bd_K, pc2bd_I, TransformCrossSectionMatrix
import logging
logger = logging.getLogger("wisdem/weis")


class RunPreComp(ExplicitComponent):
    # Openmdao component to run precomp and generate the elastic properties of a wind turbine blade
    def initialize(self):
        self.options.declare("modeling_options")
        self.options.declare("opt_options")

    def setup(self):
        rotorse_options = self.options["modeling_options"]["WISDEM"]["RotorSE"]
        self.n_span = n_span = rotorse_options["n_span"]
        self.n_webs = n_webs = rotorse_options["n_webs"]
        self.n_layers = n_layers = rotorse_options["n_layers"]
        self.n_xy = n_xy = rotorse_options["n_xy"]  # Number of coordinate points to describe the airfoil geometry
        mat_init_options = self.options["modeling_options"]["materials"]
        self.n_mat = n_mat = mat_init_options["n_mat"]
        self.verbosity = self.options["modeling_options"]["General"]["verbosity"]

        self.te_ss_var = rotorse_options["te_ss"]
        self.te_ps_var = rotorse_options["te_ps"]
        self.spar_cap_ss_var = rotorse_options["spar_cap_ss"]
        self.spar_cap_ps_var = rotorse_options["spar_cap_ps"]

        # Outer geometry
        self.add_input(
            "r",
            val=np.zeros(n_span),
            units="m",
            desc="radial locations where blade is defined (should be increasing and not go all the way to hub or tip)",
        )
        self.add_input(
            "theta",
            val=np.zeros(n_span),
            units="deg",
            desc="Twist angle at each section (positive decreases angle of attack)",
        )
        self.add_input("chord", val=np.zeros(n_span), units="m", desc="chord length at each section")
        self.add_input(
            "pitch_axis",
            val=np.zeros(n_span),
            desc="1D array of the chordwise position of the pitch axis (0-LE, 1-TE), defined along blade span.",
        )
        self.add_input("precurve", val=np.zeros(n_span), units="m", desc="precurve at each section")
        self.add_input("presweep", val=np.zeros(n_span), units="m", desc="presweep at each section")
        self.add_input(
            "coord_xy_interp",
            val=np.zeros((n_span, n_xy, 2)),
            desc="3D array of the non-dimensional x and y airfoil coordinates of the airfoils interpolated along span for n_span stations.",
        )

        # Rotor configuration
        self.add_input(
            "uptilt", val=0.0, units="deg", desc="Nacelle uptilt angle. A standard machine has positive values."
        )
        self.add_discrete_input("n_blades", val=3, desc="Number of blades of the rotor.")

        # Inner structure
        self.add_input(
            "web_start_nd",
            val=np.zeros((n_webs, n_span)),
            desc="2D array of the non-dimensional start point defined along the outer profile of a web. The TE suction side is 0, the TE pressure side is 1. The first dimension represents each web, the second dimension represents each entry along blade span.",
        )
        self.add_input(
            "web_end_nd",
            val=np.zeros((n_webs, n_span)),
            desc="2D array of the non-dimensional end point defined along the outer profile of a web. The TE suction side is 0, the TE pressure side is 1. The first dimension represents each web, the second dimension represents each entry along blade span.",
        )
        self.add_input(
            "layer_web",
            val=np.zeros(n_layers),
            desc="1D array of the web id the layer is associated to. If the layer is on the outer profile, this entry can simply stay equal to 0.",
        )
        self.add_discrete_input(
            "definition_layer",
            val=np.zeros(n_layers),
            desc="1D array of flags identifying how layers are specified in the yaml. 1) all around (skin, paint, ) 2) offset+rotation twist+width (spar caps) 3) offset+user defined rotation+width 4) midpoint TE+width (TE reinf) 5) midpoint LE+width (LE reinf) 6) layer position fixed to other layer (core fillers) 7) start and width 8) end and width 9) start and end nd 10) web layer",
        )
        self.add_input(
            "layer_thickness",
            val=np.zeros((n_layers, n_span)),
            units="m",
            desc="2D array of the thickness of the layers of the blade structure. The first dimension represents each layer, the second dimension represents each entry along blade span.",
        )
        self.add_input(
            "layer_start_nd",
            val=np.zeros((n_layers, n_span)),
            desc="2D array of the non-dimensional start point defined along the outer profile of a layer. The TE suction side is 0, the TE pressure side is 1. The first dimension represents each layer, the second dimension represents each entry along blade span.",
        )
        self.add_input(
            "layer_end_nd",
            val=np.zeros((n_layers, n_span)),
            desc="2D array of the non-dimensional end point defined along the outer profile of a layer. The TE suction side is 0, the TE pressure side is 1. The first dimension represents each layer, the second dimension represents each entry along blade span.",
        )
        self.add_input(
            "fiber_orientation",
            val=np.zeros((n_layers, n_span)),
            units="deg",
            desc="2D array of the orientation of the layers of the blade structure. The first dimension represents each layer, the second dimension represents each entry along blade span.",
        )

        # Materials
        self.add_discrete_input("mat_name", val=n_mat * [""], desc="1D array of names of materials.")
        self.add_discrete_input(
            "orth",
            val=np.zeros(n_mat),
            desc="1D array of flags to set whether a material is isotropic (0) or orthtropic (1). Each entry represents a material.",
        )
        self.add_input(
            "E",
            val=np.zeros([n_mat, 3]),
            units="Pa",
            desc="2D array of the Youngs moduli of the materials. Each row represents a material, the three columns represent E11, E22 and E33.",
        )
        self.add_input(
            "G",
            val=np.zeros([n_mat, 3]),
            units="Pa",
            desc="2D array of the shear moduli of the materials. Each row represents a material, the three columns represent G12, G13 and G23.",
        )
        self.add_input(
            "nu",
            val=np.zeros([n_mat, 3]),
            desc="2D array of the Poisson ratio of the materials. Each row represents a material, the three columns represent nu12, nu13 and nu23.",
        )
        self.add_input(
            "rho",
            val=np.zeros(n_mat),
            units="kg/m**3",
            desc="1D array of the density of the materials. For composites, this is the density of the laminate.",
        )

        self.add_input(
            "joint_position",
            val=0.0,
            desc="Spanwise position of the segmentation joint.",
        )
        self.add_input("joint_mass", val=0.0, units="kg", desc="Mass of the joint.")

        # Outputs - Distributed beam properties
        self.add_output("z", val=np.zeros(n_span), units="m", desc="locations of properties along beam")
        self.add_output("A", val=np.zeros(n_span), units="m**2", desc="cross sectional area")
        self.add_output("EA", val=np.zeros(n_span), units="N", desc="axial stiffness")
        self.add_output(
            "EIxx",
            val=np.zeros(n_span),
            units="N*m**2",
            desc="Section lag (edgewise) bending stiffness about the XE axis",
        )
        self.add_output(
            "EIyy",
            val=np.zeros(n_span),
            units="N*m**2",
            desc="Section flap bending stiffness about the YE axis",
        )
        self.add_output("EIxy", val=np.zeros(n_span), units="N*m**2", desc="Coupled flap-lag stiffness with respect to the XE-YE frame")
        self.add_output("EA_EIxx", val=np.zeros(n_span), units="N*m", desc="Coupled axial-lag stiffness with respect to the XE-YE frame")
        self.add_output("EA_EIyy", val=np.zeros(n_span), units="N*m", desc="Coupled axial-flap stiffness with respect to the XE-YE frame")
        self.add_output("EIxx_GJ", val=np.zeros(n_span), units="N*m**2", desc="Coupled lag-torsion stiffness with respect to the XE-YE frame")
        self.add_output("EIyy_GJ", val=np.zeros(n_span), units="N*m**2", desc="Coupled flap-torsion stiffness with respect to the XE-YE frame ")
        self.add_output("EA_GJ", val=np.zeros(n_span), units="N*m", desc="Coupled axial-torsion stiffness")
        self.add_output(
            "GJ",
            val=np.zeros(n_span),
            units="N*m**2",
            desc="Section torsional stiffness with respect to the XE-YE frame",
        )
        self.add_output("rhoA", val=np.zeros(n_span), units="kg/m", desc="Section mass per unit length")
        self.add_output("rhoJ", val=np.zeros(n_span), units="kg*m", desc="polar mass moment of inertia per unit length")
        self.add_output(
            "Tw_iner",
            val=np.zeros(n_span),
            units="deg",
            desc="Orientation of the section principal inertia axes with respect the blade reference plane",
        )
        self.add_output(
            "x_tc",
            val=np.zeros(n_span),
            units="m",
            desc="X-coordinate of the tension-center offset with respect to the XR-YR axes",
        )
        self.add_output(
            "y_tc",
            val=np.zeros(n_span),
            units="m",
            desc="Chordwise offset of the section tension-center with respect to the XR-YR axes",
        )
        self.add_output(
            "x_sc",
            val=np.zeros(n_span),
            units="m",
            desc="X-coordinate of the shear-center offset with respect to the XR-YR axes",
        )
        self.add_output(
            "y_sc",
            val=np.zeros(n_span),
            units="m",
            desc="Chordwise offset of the section shear-center with respect to the reference frame, XR-YR",
        )
        self.add_output(
            "x_cg",
            val=np.zeros(n_span),
            units="m",
            desc="X-coordinate of the center-of-mass offset with respect to the XR-YR axes",
        )
        self.add_output(
            "y_cg",
            val=np.zeros(n_span),
            units="m",
            desc="Chordwise offset of the section center of mass with respect to the XR-YR axes",
        )
        self.add_output(
            "flap_iner",
            val=np.zeros(n_span),
            units="kg/m",
            desc="Section flap inertia about the Y_G axis per unit length.",
        )
        self.add_output(
            "edge_iner",
            val=np.zeros(n_span),
            units="kg/m",
            desc="Section lag inertia about the X_G axis per unit length",
        )
        # self.add_output('eps_crit_spar',    val=np.zeros(n_span), desc='critical strain in spar from panel buckling calculation')
        # self.add_output('eps_crit_te',      val=np.zeros(n_span), desc='critical strain in trailing-edge panels from panel buckling calculation')
        self.add_output(
            "xu_spar",
            val=np.zeros(n_span),
            desc="x-position of midpoint of spar cap on upper surface for strain calculation",
        )
        self.add_output(
            "xl_spar",
            val=np.zeros(n_span),
            desc="x-position of midpoint of spar cap on lower surface for strain calculation",
        )
        self.add_output(
            "yu_spar",
            val=np.zeros(n_span),
            desc="y-position of midpoint of spar cap on upper surface for strain calculation",
        )
        self.add_output(
            "yl_spar",
            val=np.zeros(n_span),
            desc="y-position of midpoint of spar cap on lower surface for strain calculation",
        )
        self.add_output(
            "xu_te",
            val=np.zeros(n_span),
            desc="x-position of midpoint of trailing-edge panel on upper surface for strain calculation",
        )
        self.add_output(
            "xl_te",
            val=np.zeros(n_span),
            desc="x-position of midpoint of trailing-edge panel on lower surface for strain calculation",
        )
        self.add_output(
            "yu_te",
            val=np.zeros(n_span),
            desc="y-position of midpoint of trailing-edge panel on upper surface for strain calculation",
        )
        self.add_output(
            "yl_te",
            val=np.zeros(n_span),
            desc="y-position of midpoint of trailing-edge panel on lower surface for strain calculation",
        )


        self.add_output(
            "sc_ss_mats",
            val=np.zeros((n_span, n_mat)),
            desc="spar cap, suction side,  boolean of materials in each composite layer spanwise, passed as floats for differentiablity, used for Fatigue Analysis",
        )
        self.add_output(
            "sc_ps_mats",
            val=np.zeros((n_span, n_mat)),
            desc="spar cap, pressure side, boolean of materials in each composite layer spanwise, passed as floats for differentiablity, used for Fatigue Analysis",
        )
        self.add_output(
            "te_ss_mats",
            val=np.zeros((n_span, n_mat)),
            desc="trailing edge reinforcement, suction side,  boolean of materials in each composite layer spanwise, passed as floats for differentiablity, used for Fatigue Analysis",
        )
        self.add_output(
            "te_ps_mats",
            val=np.zeros((n_span, n_mat)),
            desc="trailing edge reinforcement, pressure side, boolean of materials in each composite layer spanwise, passed as floats for differentiablity, used for Fatigue Analysis",
        )

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        ##############################
        def region_stacking(
            i,
            idx,
            start_nd_arc,
            end_nd_arc,
            layer_name,
            layer_thickness,
            fiber_orientation,
            layer_mat,
            material_dict,
            materials,
            region_loc,
        ):
            # Receive start and end of composite sections chordwise, find which composites layers are in each
            # chordwise regions, generate the precomp composite class instance

            # error handling to makes sure there were no numeric errors causing values very close too, but not exactly, 0 or 1
            start_nd_arc = [
                0.0 if start_nd_arci != 0.0 and np.isclose(start_nd_arci, 0.0) else start_nd_arci
                for start_nd_arci in start_nd_arc
            ]
            end_nd_arc = [
                0.0 if end_nd_arci != 0.0 and np.isclose(end_nd_arci, 0.0) else end_nd_arci
                for end_nd_arci in end_nd_arc
            ]
            start_nd_arc = [
                1.0 if start_nd_arci != 1.0 and np.isclose(start_nd_arci, 1.0) else start_nd_arci
                for start_nd_arci in start_nd_arc
            ]
            end_nd_arc = [
                1.0 if end_nd_arci != 1.0 and np.isclose(end_nd_arci, 1.0) else end_nd_arci
                for end_nd_arci in end_nd_arc
            ]

            # region end points
            dp = sorted(list(set(start_nd_arc + end_nd_arc)))

            # initialize
            n_plies = []
            thk = []
            theta = []
            mat_idx = []

            # loop through division points, find what layers make up the stack between those bounds
            for i_reg, (dp0, dp1) in enumerate(zip(dp[0:-1], dp[1:])):
                n_pliesi = []
                thki = []
                thetai = []
                mati = []
                for i_sec, start_nd_arci, end_nd_arci in zip(idx, start_nd_arc, end_nd_arc):
                    name = layer_name[i_sec]
                    if start_nd_arci <= dp0 and end_nd_arci >= dp1:
                        if name in region_loc.keys():
                            if region_loc[name][i] == None:
                                region_loc[name][i] = [i_reg]
                            else:
                                region_loc[name][i].append(i_reg)

                        n_pliesi.append(1.0)
                        thki.append(layer_thickness[i_sec])
                        if fiber_orientation[i_sec] == None:
                            thetai.append(0.0)
                        else:
                            thetai.append(fiber_orientation[i_sec])
                        mati.append(material_dict[layer_mat[i_sec]])

                n_plies.append(np.array(n_pliesi))
                thk.append(np.array(thki))
                theta.append(np.array(thetai))
                mat_idx.append(np.array(mati))

            # print('----------------------')
            # print('dp', dp)
            # print('n_plies', n_plies)
            # print('thk', thk)
            # print('theta', theta)
            # print('mat_idx', mat_idx)
            # print('materials', materials)

            sec = CompositeSection(dp, n_plies, thk, theta, mat_idx, materials)
            return sec, region_loc
            ##############################

        def web_stacking(
            i,
            web_idx,
            web_start_nd_arc,
            web_end_nd_arc,
            layer_thickness,
            fiber_orientation,
            layer_mat,
            material_dict,
            materials,
            flatback,
            upperCSi,
        ):
            dp = []
            n_plies = []
            thk = []
            theta = []
            mat_idx = []

            if len(web_idx) > 0:
                dp = np.mean((np.abs(web_start_nd_arc), np.abs(web_start_nd_arc)), axis=0).tolist()

                dp_all = [
                    [-1.0 * start_nd_arci, -1.0 * end_nd_arci]
                    for start_nd_arci, end_nd_arci in zip(web_start_nd_arc, web_end_nd_arc)
                ]
                web_dp, web_ids = np.unique(dp_all, axis=0, return_inverse=True)
                for webi in np.unique(web_ids):
                    # store variable values (thickness, orientation, material) for layers that make up each web, based on the mapping array web_ids
                    n_pliesi = [1.0 for i_reg, web_idi in zip(web_idx, web_ids) if web_idi == webi]
                    thki = [layer_thickness[i_reg] for i_reg, web_idi in zip(web_idx, web_ids) if web_idi == webi]
                    thetai = [fiber_orientation[i_reg] for i_reg, web_idi in zip(web_idx, web_ids) if web_idi == webi]
                    thetai = [0.0 if theta_ij == None else theta_ij for theta_ij in thetai]
                    mati = [
                        material_dict[layer_mat[i_reg]] for i_reg, web_idi in zip(web_idx, web_ids) if web_idi == webi
                    ]

                    n_plies.append(np.array(n_pliesi))
                    thk.append(np.array(thki))
                    theta.append(np.array(thetai))
                    mat_idx.append(np.array(mati))

            if flatback:
                dp.append(1.0)
                n_plies.append(upperCSi.n_plies[-1])
                thk.append(upperCSi.t[-1])
                theta.append(upperCSi.theta[-1])
                mat_idx.append(upperCSi.mat_idx[-1])

            dp_out = sorted(list(set(dp)))

            sec = CompositeSection(dp_out, n_plies, thk, theta, mat_idx, materials)
            return sec
            ##############################

        layer_name = self.options["modeling_options"]["WISDEM"]["RotorSE"]["layer_name"]
        layer_mat = self.options["modeling_options"]["WISDEM"]["RotorSE"]["layer_mat"]

        upperCS = [None] * self.n_span
        lowerCS = [None] * self.n_span
        websCS = [None] * self.n_span
        profile = [None] * self.n_span
        chord = inputs["chord"]
        area = np.zeros_like(chord)
        region_loc_vars = [self.te_ss_var, self.te_ps_var, self.spar_cap_ss_var, self.spar_cap_ps_var]

        region_loc_ss = {}  # track precomp regions for user selected composite layers
        region_loc_ps = {}
        for var in region_loc_vars:
            region_loc_ss[var] = [None] * self.n_span
            region_loc_ps[var] = [None] * self.n_span

        ## Materials
        material_dict = {}
        materials = []
        for i_mat in range(self.n_mat):
            materials.append(
                Orthotropic2DMaterial(
                    inputs["E"][i_mat, 0],
                    inputs["E"][i_mat, 1],
                    inputs["G"][i_mat, 0],
                    inputs["nu"][i_mat, 0],
                    inputs["rho"][i_mat],
                    discrete_inputs["mat_name"][i_mat],
                )
            )
            material_dict[discrete_inputs["mat_name"][i_mat]] = i_mat

        ## Spanwise
        for i in range(self.n_span):
            # time0 = time.time()

            ## Profiles
            # rotate
            profile_i = inputs["coord_xy_interp"][i, :, :]
            profile_i_rot = profile_i
            # Trying to ensure all shear webs are in perpendicular orientation,
            # but this probably doesn't work as robustly as hoped.
            #profile_i_rot = np.column_stack(
            #    rotate(inputs["pitch_axis"][i], 0.0, profile_i[:, 0], profile_i[:, 1], np.radians(inputs["theta"][i]))
            #)

            # import matplotlib.pyplot as plt
            # plt.plot(profile_i[:,0], profile_i[:,1])
            # plt.plot(profile_i_rot[:,0], profile_i_rot[:,1])
            # plt.axis('equal')
            # plt.title(i)
            # plt.show()

            # normalize
            profile_i_rot[:, 0] -= min(profile_i_rot[:, 0])
            profile_i_rot = profile_i_rot / max(profile_i_rot[:, 0])

            profile_i_rot_precomp = copy.copy(profile_i_rot)
            idx_s = 0
            idx_le_precomp = np.argmax(profile_i_rot_precomp[:, 0])
            if idx_le_precomp != 0:
                if profile_i_rot_precomp[0, 0] == profile_i_rot_precomp[-1, 0]:
                    idx_s = 1
                profile_i_rot_precomp = np.row_stack(
                    (profile_i_rot_precomp[idx_le_precomp:], profile_i_rot_precomp[idx_s:idx_le_precomp, :])
                )
            profile_i_rot_precomp[:, 1] -= profile_i_rot_precomp[np.argmin(profile_i_rot_precomp[:, 0]), 1]

            # # renormalize
            profile_i_rot_precomp[:, 0] -= min(profile_i_rot_precomp[:, 0])
            profile_i_rot_precomp = profile_i_rot_precomp / max(profile_i_rot_precomp[:, 0])

            if profile_i_rot_precomp[-1, 0] != 1.0:
                profile_i_rot_precomp = np.row_stack((profile_i_rot_precomp, profile_i_rot_precomp[0, :]))

            # 'web' at trailing edge needed for flatback airfoils
            if (
                profile_i_rot_precomp[0, 1] != profile_i_rot_precomp[-1, 1]
                and profile_i_rot_precomp[0, 0] == profile_i_rot_precomp[-1, 0]
            ):
                flatback = True
            else:
                flatback = False

            profile[i] = Profile.initWithTEtoTEdata(profile_i_rot_precomp[:, 0], profile_i_rot_precomp[:, 1])

            # import matplotlib.pyplot as plt
            # plt.plot(profile_i_rot_precomp[:,0], profile_i_rot_precomp[:,1])
            # plt.axis('equal')
            # plt.title(i)
            # plt.show()

            idx_le = np.argmin(profile_i_rot[:, 0])

            profile_i_arc = arc_length(profile_i_rot)
            arc_L = profile_i_arc[-1]
            profile_i_arc /= arc_L
            arc_L_m = arc_L * chord[i]

            loc_LE = profile_i_arc[idx_le]
            len_PS = 1.0 - loc_LE

            ## Composites
            ss_idx = []
            ss_start_nd_arc = []
            ss_end_nd_arc = []
            ps_idx = []
            ps_start_nd_arc = []
            ps_end_nd_arc = []
            web_start_nd_arc = []
            web_end_nd_arc = []
            web_idx = []

            # Determine spanwise composite layer elements that are non-zero at this spanwise location,
            # determine their chord-wise start and end location on the pressure and suctions side

            spline_arc2xnd = PchipInterpolator(profile_i_arc, profile_i_rot[:, 0])

            # time1 = time.time()
            for idx_sec in range(self.n_layers):
                if discrete_inputs["definition_layer"][idx_sec] != 10:
                    if inputs["layer_thickness"][idx_sec, i] > 1.0e-6:
                        area[i] += arc_L_m * (inputs["layer_end_nd"][idx_sec, i] - 
                                              inputs["layer_start_nd"][idx_sec, i]) * (
                                              inputs["layer_thickness"][idx_sec, i])
                        if inputs["layer_start_nd"][idx_sec, i] < loc_LE or inputs["layer_end_nd"][idx_sec, i] < loc_LE:
                            ss_idx.append(idx_sec)
                            if inputs["layer_start_nd"][idx_sec, i] < loc_LE:
                                # ss_start_nd_arc.append(sec['start_nd_arc']['values'][i])
                                ss_end_nd_arc_temp = float(spline_arc2xnd(inputs["layer_start_nd"][idx_sec, i]))
                                if ss_end_nd_arc_temp > 1:
                                    logger.debug(
                                        "Error in the definition of material "
                                        + layer_name[idx_sec]
                                        + ". It cannot fit in the section number "
                                        + str(i)
                                        + " at span location "
                                        + str(inputs["r"][i] / inputs["r"][-1] * 100.0)
                                        + " %. Variable ss_end_nd_arc_temp was equal "
                                        + " to "
                                        + str(ss_end_nd_arc_temp)
                                        + " and is not set to 1"
                                    )
                                    ss_end_nd_arc_temp = 1.
                                if ss_end_nd_arc_temp < 0:
                                    logger.debug(
                                        "Error in the definition of material "
                                        + layer_name[idx_sec]
                                        + ". It cannot fit in the section number "
                                        + str(i)
                                        + " at span location "
                                        + str(inputs["r"][i] / inputs["r"][-1] * 100.0)
                                        + " %. Variable ss_end_nd_arc_temp was equal "
                                        + " to "
                                        + str(ss_end_nd_arc_temp)
                                        + " and is not set to 0"
                                    )
                                    ss_end_nd_arc_temp = 0.
                                if ss_end_nd_arc_temp == profile_i_rot[0, 0] and profile_i_rot[0, 0] != 1.0:
                                    ss_end_nd_arc_temp = 1.0
                                ss_end_nd_arc.append(ss_end_nd_arc_temp)
                            else:
                                ss_end_nd_arc.append(1.0)
                            # ss_end_nd_arc.append(min(sec['end_nd_arc']['values'][i], loc_LE)/loc_LE)
                            if inputs["layer_end_nd"][idx_sec, i] < loc_LE:
                                ss_start_nd_arc.append(float(spline_arc2xnd(inputs["layer_end_nd"][idx_sec, i])))
                            else:
                                ss_start_nd_arc.append(0.0)

                        if inputs["layer_start_nd"][idx_sec, i] > loc_LE or inputs["layer_end_nd"][idx_sec, i] > loc_LE:
                            ps_idx.append(idx_sec)
                            # ps_start_nd_arc.append((max(sec['start_nd_arc']['values'][i], loc_LE)-loc_LE)/len_PS)
                            # ps_end_nd_arc.append((min(sec['end_nd_arc']['values'][i], 1.)-loc_LE)/len_PS)

                            if (
                                inputs["layer_start_nd"][idx_sec, i] > loc_LE
                                and inputs["layer_end_nd"][idx_sec, i] < loc_LE
                            ):
                                # ps_start_nd_arc.append(float(remap2grid(profile_i_arc, profile_i_rot[:,0], sec['start_nd_arc']['values'][i])))
                                ps_end_nd_arc.append(1.0)
                            else:
                                ps_end_nd_arc_temp = float(spline_arc2xnd(inputs["layer_end_nd"][idx_sec, i]))
                                if (
                                    np.isclose(ps_end_nd_arc_temp, profile_i_rot[-1, 0], atol=1.0e-2)
                                    and profile_i_rot[-1, 0] != 1.0
                                ):
                                    ps_end_nd_arc_temp = 1.0
                                if ps_end_nd_arc_temp > 1.0:
                                    ps_end_nd_arc_temp = 1.0
                                ps_end_nd_arc.append(ps_end_nd_arc_temp)
                            if inputs["layer_start_nd"][idx_sec, i] < loc_LE:
                                ps_start_nd_arc.append(0.0)
                            else:
                                ps_start_nd_arc.append(float(spline_arc2xnd(inputs["layer_start_nd"][idx_sec, i])))
                else:
                    target_idx = inputs["layer_web"][idx_sec] - 1

                    if inputs["layer_thickness"][idx_sec, i] > 1.0e-6:
                        web_idx.append(idx_sec)

                        web_start_nd = inputs["web_start_nd"][int(target_idx), i]
                        web_end_nd = inputs["web_end_nd"][int(target_idx), i]

                        start_nd_arc = float(spline_arc2xnd(web_start_nd))
                        end_nd_arc = float(spline_arc2xnd(web_end_nd))

                        web_start_nd_arc.append(start_nd_arc)
                        web_end_nd_arc.append(end_nd_arc)

                        # Compute height the webs along span
                        id_start = np.argmin(abs(profile_i_arc - web_start_nd))
                        id_end = np.argmin(abs(profile_i_arc - web_end_nd))
                        web_height = np.sqrt((profile_i[id_start, 0] - profile_i[id_end, 0])**2 +
                                             (profile_i[id_start, 1] - profile_i[id_end, 1])**2) * (
                                             chord[i])

                        area[i] += web_height * inputs["layer_thickness"][idx_sec, i]

            # time1 = time.time() - time1
            # print(time1)

            # cap layer starts and ends within 0 and 1
            ss_start_nd_arc = [max(0, min(1, value)) for value in ss_start_nd_arc]
            ss_end_nd_arc = [max(0, min(1, value)) for value in ss_end_nd_arc]
            # generate the Precomp composite stacks for chordwise regions
            upperCS[i], region_loc_ss = region_stacking(
                i,
                ss_idx,
                ss_start_nd_arc,
                ss_end_nd_arc,
                layer_name,
                inputs["layer_thickness"][:, i],
                inputs["fiber_orientation"][:, i],
                layer_mat,
                material_dict,
                materials,
                region_loc_ss,
            )
            lowerCS[i], region_loc_ps = region_stacking(
                i,
                ps_idx,
                ps_start_nd_arc,
                ps_end_nd_arc,
                layer_name,
                inputs["layer_thickness"][:, i],
                inputs["fiber_orientation"][:, i],
                layer_mat,
                material_dict,
                materials,
                region_loc_ps,
            )
            if len(web_idx) > 0 or flatback:
                websCS[i] = web_stacking(
                    i,
                    web_idx,
                    web_start_nd_arc,
                    web_end_nd_arc,
                    inputs["layer_thickness"][:, i],
                    inputs["fiber_orientation"][:, i],
                    layer_mat,
                    material_dict,
                    materials,
                    flatback,
                    upperCS[i],
                )
            else:
                websCS[i] = CompositeSection([], [], [], [], [], [])

        sector_idx_spar_cap_ss = [
            None if regs == None else regs[int(len(regs) / 2)] for regs in region_loc_ss[self.spar_cap_ss_var]
        ]
        sector_idx_spar_cap_ps = [
            None if regs == None else regs[int(len(regs) / 2)] for regs in region_loc_ps[self.spar_cap_ps_var]
        ]
        sector_idx_te_ss = [
            None if regs == None else regs[int(len(regs) / 2)] for regs in region_loc_ss[self.te_ss_var]
        ]
        sector_idx_te_ps = [
            None if regs == None else regs[int(len(regs) / 2)] for regs in region_loc_ps[self.te_ps_var]
        ]

        # Get Beam Properties
        beam = PreComp(
            inputs["r"],
            inputs["chord"],
            inputs["theta"],
            inputs["pitch_axis"],
            inputs["precurve"],
            inputs["presweep"],
            profile,
            materials,
            upperCS,
            lowerCS,
            websCS,
            sector_idx_spar_cap_ps,
            sector_idx_spar_cap_ss,
            sector_idx_te_ps,
            sector_idx_te_ss,
        )
        (
            EIxx,
            EIyy,
            GJ,
            EA,
            EIxy,
            EA_EIxx,
            EA_EIyy,
            EIxx_GJ,
            EIyy_GJ,
            EA_GJ,
            rhoA,
            _,
            rhoJ,
            Tw_iner,
            flap_iner,
            edge_iner,
            x_tc,
            y_tc,
            x_sc,
            y_sc,
            x_cg,
            y_cg,
        ) = beam.sectionProperties()

        # outputs['eps_crit_spar'] = beam.panelBucklingStrain(sector_idx_spar_cap_ss)
        # outputs['eps_crit_te'] = beam.panelBucklingStrain(sector_idx_te_ss)

        xu_spar, xl_spar, yu_spar, yl_spar = beam.criticalStrainLocations(
            sector_idx_spar_cap_ss, sector_idx_spar_cap_ps
        )
        xu_te, xl_te, yu_te, yl_te = beam.criticalStrainLocations(sector_idx_te_ss, sector_idx_te_ps)

        # Store what materials make up the composites for SC/TE
        for i in range(self.n_span):
            for j in range(self.n_mat):
                if sector_idx_spar_cap_ss[i]:
                    if j in upperCS[i].mat_idx[sector_idx_spar_cap_ss[i]]:
                        outputs["sc_ss_mats"][i, j] = 1.0
                if sector_idx_spar_cap_ps[i]:
                    if j in lowerCS[i].mat_idx[sector_idx_spar_cap_ps[i]]:
                        outputs["sc_ps_mats"][i, j] = 1.0
                if sector_idx_te_ss[i]:
                    if j in upperCS[i].mat_idx[sector_idx_te_ss[i]]:
                        outputs["te_ss_mats"][i, j] = 1.0
                if sector_idx_te_ps[i]:
                    if j in lowerCS[i].mat_idx[sector_idx_te_ps[i]]:
                        outputs["te_ps_mats"][i, j] = 1.0
        rhoA_joint = copy.copy(rhoA)
        if inputs["joint_mass"] > 0.0:
            s = (inputs["r"] - inputs["r"][0]) / (inputs["r"][-1] - inputs["r"][0])
            id_station = np.argmin(abs(inputs["joint_position"] - s))
            span = np.average(
                [
                    inputs["r"][id_station] - inputs["r"][id_station - 1],
                    inputs["r"][id_station + 1] - inputs["r"][id_station],
                ]
            )
            rhoA_joint[id_station] += inputs["joint_mass"][0] / span

        outputs["z"] = inputs["r"]
        outputs["EIxx"] = EIxx
        outputs["EIyy"] = EIyy
        outputs["GJ"] = GJ
        outputs["EA"] = EA
        outputs["EIxy"] = EIxy
        outputs["EA_EIxx"] = EA_EIxx
        outputs["EA_EIyy"] = EA_EIyy
        outputs["EIxx_GJ"] = EIxx_GJ
        outputs["EIyy_GJ"] = EIyy_GJ
        outputs["EA_GJ"] = EA_GJ
        outputs["rhoA"] = rhoA_joint
        outputs["A"] = area
        outputs["rhoJ"] = rhoJ
        outputs["Tw_iner"] = Tw_iner
        outputs["flap_iner"] = flap_iner
        outputs["edge_iner"] = edge_iner

        outputs["x_tc"] = x_tc
        outputs["y_tc"] = y_tc
        outputs["x_sc"] = x_sc
        outputs["y_sc"] = y_sc
        outputs["x_cg"] = x_cg
        outputs["y_cg"] = y_cg

        outputs["xu_spar"] = xu_spar
        outputs["xl_spar"] = xl_spar
        outputs["yu_spar"] = yu_spar
        outputs["yl_spar"] = yl_spar
        outputs["xu_te"] = xu_te
        outputs["xl_te"] = xl_te
        outputs["yu_te"] = yu_te
        outputs["yl_te"] = yl_te



class TotalBladeProperties(ExplicitComponent):
    def initialize(self):
        self.options.declare("modeling_options")
        self.options.declare("opt_options")

    def setup(self):
        rotorse_options = self.options["modeling_options"]["WISDEM"]["RotorSE"]
        self.n_span = n_span = rotorse_options["n_span"]

        # Inputs
        self.add_input(
            "r",
            val=np.zeros(n_span),
            units="m",
            desc="radial locations where blade is defined (should be increasing and not go all the way to hub or tip)",
        )
        self.add_input("rhoA", val=np.zeros(n_span), units="kg/m", desc="mass per unit length")
        self.add_discrete_input("n_blades", val=3, desc="Number of blades of the rotor.")

        # Outputs - Overall beam properties
        self.add_output("blade_mass", val=0.0, units="kg", desc="mass of one blade")
        self.add_output("blade_span_cg", val=0.0, units="m", desc="Distance along the blade span for its center of gravity")
        self.add_output(
            "blade_moment_of_inertia", val=0.0, units="kg*m**2", desc="mass moment of inertia of blade about hub"
        )
        self.add_output("mass_all_blades", val=0.0, units="kg", desc="mass of all blades")
        self.add_output(
            "I_all_blades",
            shape=6,
            units="kg*m**2",
            desc="mass moments of inertia of all blades in hub c.s. order:Ixx, Iyy, Izz, Ixy, Ixz, Iyz",
        )

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        rhoA_joint = inputs["rhoA"]
        blade_mass = np.trapz(rhoA_joint, inputs["r"])
        blade_span_cg = np.trapz(rhoA_joint * inputs["r"], inputs["r"]) / blade_mass
        blade_moment_of_inertia = np.trapz(rhoA_joint * inputs["r"] ** 2.0, inputs["r"])
        # tilt = inputs["uptilt"]
        n_blades = discrete_inputs["n_blades"]
        mass_all_blades = n_blades * blade_mass
        Ibeam = n_blades * blade_moment_of_inertia
        Ixx = Ibeam
        Iyy = Ibeam / 2.0  # azimuthal average for 2 blades, exact for 3+
        Izz = Ibeam / 2.0
        Ixy = 0.0
        Ixz = 0.0
        Iyz = 0.0  # azimuthal average for 2 blades, exact for 3+
        I_all_blades = np.r_[Ixx, Iyy, Izz, Ixy, Ixz, Iyz]

        outputs["blade_mass"] = blade_mass
        outputs["blade_span_cg"] = blade_span_cg
        outputs["blade_moment_of_inertia"] = blade_moment_of_inertia
        outputs["mass_all_blades"] = mass_all_blades
        outputs["I_all_blades"] = I_all_blades
                
class generate_KI(ExplicitComponent):

    def initialize(self):
        self.options.declare('modeling_options')

    def setup(self):
        
        modopt = self.options['modeling_options']
        rotorse_options  = modopt['WISDEM']['RotorSE']
        self.n_span = n_span = rotorse_options['n_span']

        self.add_input(
            "EA", 
            val=np.zeros(n_span), 
            units="N", 
            desc="Axial stiffness at the elastic center, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "EIxx",
            val=np.zeros(n_span),
            units="N*m**2",
            desc="Section lag (edgewise) bending stiffness about the XE axis, using the convention of WISDEM solver PreComp.",
        )
        self.add_input("EIyy",
            val=np.zeros(n_span),
            units="N*m**2",
            desc="Section flap bending stiffness about the YE axis, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "EIxy", 
            val=np.zeros(n_span), 
            units="N*m**2", 
            desc="Coupled flap-lag stiffness with respect to the XE-YE frame, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "EA_EIxx", 
            val=np.zeros(n_span), 
            units="N*m", 
            desc="Coupled axial-lag stiffness with respect to the XE-YE frame, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "EA_EIyy", 
            val=np.zeros(n_span), 
            units="N*m", 
            desc="Coupled axial-flap stiffness with respect to the XE-YE frame, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "EIxx_GJ", 
            val=np.zeros(n_span), 
            units="N*m**2", 
            desc="Coupled lag-torsion stiffness with respect to the XE-YE frame, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "EIyy_GJ", 
            val=np.zeros(n_span), 
            units="N*m**2", 
            desc="Coupled flap-torsion stiffness with respect to the XE-YE frame, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "EA_GJ", 
            val=np.zeros(n_span), 
            units="N*m", 
            desc="Coupled axial-torsion stiffness with respect to the XE-YE frame, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "GJ",
            val=np.zeros(n_span),
            units="N*m**2",
            desc="Section torsion stiffness at the elastic center, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "rhoA", 
            val=np.zeros(n_span), 
            units="kg/m", 
            desc="Section mass per unit length, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "rhoJ", 
            val=np.zeros(n_span), 
            units="kg*m", 
            desc="polar mass moment of inertia per unit length, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "Tw_iner",
            val=np.zeros(n_span),
            units="deg",
            desc="Orientation of the section principal inertia axes with respect the blade reference plane, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "x_tc",
            val=np.zeros(n_span),
            units="m",
            desc="X-coordinate of the tension-center offset with respect to the XR-YR axes, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "y_tc",
            val=np.zeros(n_span),
            units="m",
            desc="Chordwise offset of the section tension-center with respect to the XR-YR axes, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "x_cg",
            val=np.zeros(n_span),
            units="m",
            desc="X-coordinate of the center-of-mass offset with respect to the XR-YR axes, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "y_cg",
            val=np.zeros(n_span),
            units="m",
            desc="Chordwise offset of the section center of mass with respect to the XR-YR axes, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "flap_iner",
            val=np.zeros(n_span),
            units="kg/m",
            desc="Section flap inertia about the Y_G axis per unit length, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "edge_iner",
            val=np.zeros(n_span),
            units="kg/m",
            desc="Section lag inertia about the X_G axis per unit length, using the convention of WISDEM solver PreComp.",
        )
        self.add_input(
            "theta",
            val=np.zeros(n_span),
            units="deg",
            desc="Aerodynamic twist angle at each section (positive decreases angle of attack)",
        )

        # Outputs are 6x6 K and I matrices at the center of the windIO reference axes
        self.add_output(
            "K",
            val=np.zeros((n_span,6,6)), 
            desc="Stiffness matrix at the center of the windIO reference axes."
        )
        self.add_output(
            "I",
            val=np.zeros((n_span,6,6)), 
            desc="Inertia matrix at the center of the windIO reference axes."
        )

    def compute(self, inputs, outputs):

        # Initialize empty 6x6 K and I matrices
        K = np.zeros((self.n_span,6,6))
        I = np.zeros((self.n_span,6,6))
        EA = inputs["EA"]
        EIxx = inputs["EIxx"]
        EIyy = inputs["EIyy"]
        EIxy = inputs["EIxy"]
        GJ = inputs["GJ"]
        EA_EIxx = inputs["EA_EIxx"]
        EA_EIyy = inputs["EA_EIyy"]
        EIxx_GJ = inputs["EIxx_GJ"]
        EIyy_GJ = inputs["EIyy_GJ"]
        EA_GJ = inputs["EA_GJ"]
        rhoA = inputs["rhoA"]
        rhoJ = inputs["rhoJ"]
        Tw_iner = np.deg2rad(inputs["Tw_iner"])
        x_cg = inputs["x_cg"]
        y_cg = inputs["y_cg"]
        x_tc = inputs["x_tc"]
        y_tc = inputs["y_tc"]
        edge_iner = inputs["edge_iner"]
        flap_iner = inputs["flap_iner"]
        aero_twist =  np.deg2rad(inputs["theta"])

        for i in range(self.n_span):
            # Build stiffness matrix at the reference axis
            K[i,:,:] = pc2bd_K(
                EA[i],
                EIxx[i],
                EIyy[i],
                EIxy[i],
                EA_EIxx[i],
                EA_EIyy[i],
                EIxx_GJ[i],
                EIyy_GJ[i],
                EA_GJ[i],
                GJ[i],
                rhoJ[i],
                edge_iner[i],
                flap_iner[i],
                x_tc[i],
                y_tc[i],
                )
            # Build inertia matrix at the reference axis
            I[i,:,:] = pc2bd_I(
                rhoA[i],
                edge_iner[i],
                flap_iner[i],
                rhoJ[i],
                x_cg[i],
                y_cg[i],
                Tw_iner[i],
                aero_twist[i],
                )
        
        outputs["K"] = K
        outputs["I"] = I

class assemble_KI(ExplicitComponent):
    """"
    Assemble the stiffness and inertia matrices at the reference axis for the entire rotor blade"
    """
    def initialize(self):
        self.options.declare('modeling_options')

    def setup(self):
        modopt = self.options['modeling_options']
        rotorse_options  = modopt['WISDEM']['RotorSE']
        self.n_span = n_span = rotorse_options['n_span']

        self.add_input("K11", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K11 element of the stiffness matrix along blade span. K11 corresponds to the shear stiffness along the x axis (in a blade, x points to the trailing edge)",
                            units="N")
        self.add_input("K22", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K22 element of the stiffness matrix along blade span. K22 corresponds to the shear stiffness along the y axis (in a blade, y points to the suction side)",
                            units="N")
        self.add_input("K33", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K33 element of the stiffness matrix along blade span. K33 corresponds to the axial stiffness along the z axis (in a blade, z runs along the span and points to the tip)",
                            units="N")
        self.add_input("K44", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K44 element of the stiffness matrix along blade span. K44 corresponds to the bending stiffness around the x axis (in a blade, x points to the trailing edge and K44 corresponds to the flapwise stiffness)",
                            units="N*m**2")
        self.add_input("K55", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K55 element of the stiffness matrix along blade span. K55 corresponds to the bending stiffness around the y axis (in a blade, y points to the suction side and K55 corresponds to the edgewise stiffness)",
                            units="N*m**2")
        self.add_input("K66", 
                            val=np.zeros(n_span),  
                            desc="Distribution of K66 element of the stiffness matrix along blade span. K66 corresponds to the torsional stiffness along the z axis (in a blade, z runs along the span and points to the tip)",
                            units="N*m**2")
        self.add_input("K12", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K12 element of the stiffness matrix along blade span. K12 is a cross term between shear terms",
                            units="N")
        self.add_input("K13", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K13 element of the stiffness matrix along blade span. K13 is a cross term shear - axial",
                            units="N")
        self.add_input("K14", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K14 element of the stiffness matrix along blade span. K14 is a cross term shear - bending",
                            units="N*m**2")
        self.add_input("K15", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K15 element of the stiffness matrix along blade span. K15 is a cross term shear - bending",
                            units="N*m**2")
        self.add_input("K16", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K16 element of the stiffness matrix along blade span. K16 is a cross term shear - torsion",
                            units="N*m**2")
        self.add_input("K23", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K23 element of the stiffness matrix along blade span. K23 is a cross term shear - axial",
                            units="N*m**2")
        self.add_input("K24", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K24 element of the stiffness matrix along blade span. K24 is a cross term shear - bending",
                            units="N/m**2")
        self.add_input("K25", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K25 element of the stiffness matrix along blade span. K25 is a cross term shear - bending",
                            units="N*m**2")
        self.add_input("K26", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K26 element of the stiffness matrix along blade span. K26 is a cross term shear - torsion",
                            units="N*m**2")
        self.add_input("K34", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K34 element of the stiffness matrix along blade span. K34 is a cross term axial - bending",
                            units="N*m**2")
        self.add_input("K35", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K35 element of the stiffness matrix along blade span. K35 is a cross term axial - bending",
                            units="N*m**2")
        self.add_input("K36", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K36 element of the stiffness matrix along blade span. K36 is a cross term axial - torsion",
                            units="N*m**2")
        self.add_input("K45", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K45 element of the stiffness matrix along blade span. K45 is a cross term flapwise bending - edgewise bending",
                            units="N*m**2")
        self.add_input("K46", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K46 element of the stiffness matrix along blade span. K46 is a cross term flapwise bending - torsion",
                            units="N*m**2")
        self.add_input("K56", 
                            val=np.zeros(n_span),  
                            desc="Distribution of the K56 element of the stiffness matrix along blade span. K56 is a cross term edgewise bending - torsion",
                            units="N*m**2")
        
        # mass matrix inputs
        self.add_input("mass", val=np.zeros(n_span),  desc="Mass per unit length along the beam, expressed in kilogram per meter", units="kg/m")
        self.add_input("cm_x", val=np.zeros(n_span),  desc="Distance between the reference axis and the center of mass along the x axis", units="m")
        self.add_input("cm_y", val=np.zeros(n_span),  desc="Distance between the reference axis and the center of mass along the y axis", units="m")
        self.add_input("i_edge", val=np.zeros(n_span),  desc="Edgewise mass moment of inertia per unit span (around y axis)", units="kg*m**2")
        self.add_input("i_flap", val=np.zeros(n_span),  desc="Flapwise mass moment of inertia per unit span (around x axis)", units="kg*m**2")
        self.add_input("i_plr", val=np.zeros(n_span),  desc="Polar moment of inertia per unit span (around z axis). Please note that for beam-like structures iplr must be equal to iedge plus iflap.", units="kg*m**2")
        self.add_input("i_cp", val=np.zeros(n_span),  desc="Sectional cross-product of inertia per unit span (cross term x y)", units="kg*m**2")

        # Outputs
        self.add_output("K", val=np.zeros((n_span,6,6)), desc="Stiffness matrix at the center of the windIO reference axes.")
        self.add_output("I", val=np.zeros((n_span,6,6)), desc="Inertia matrix at the center of the windIO reference axes.")

    def compute(self, inputs, outputs):
        # Initialize empty 6x6 K and I matrices
        K = np.zeros((self.n_span,6,6))
        I = np.zeros((self.n_span,6,6))

        # Extract inputs
        K11 = inputs["K11"]
        K22 = inputs["K22"]
        K33 = inputs["K33"]
        K44 = inputs["K44"]
        K55 = inputs["K55"]
        K66 = inputs["K66"]
        K12 = inputs["K12"]
        K13 = inputs["K13"]
        K14 = inputs["K14"]
        K15 = inputs["K15"]
        K16 = inputs["K16"]
        K23 = inputs["K23"]
        K24 = inputs["K24"]
        K25 = inputs["K25"]
        K26 = inputs["K26"]
        K34 = inputs["K34"]
        K35 = inputs["K35"]
        K36 = inputs["K36"]
        K45 = inputs["K45"]
        K46 = inputs["K46"]
        K56 = inputs["K56"]

        mass = inputs['mass']
        cm_x = inputs['cm_x']
        cm_y = inputs['cm_y']
        i_edge = inputs['i_edge']
        i_flap = inputs['i_flap']
        i_plr  = inputs['i_plr']
        i_cp   = inputs['i_cp']

        
        for i in range(self.n_span):
            # Assemble stiffness matrix at the reference axis
            k_matrix = np.array([
                [K11[i], K12[i], K13[i], K14[i], K15[i], K16[i]],
                [K12[i], K22[i], K23[i], K24[i], K25[i], K26[i]],
                [K13[i], K23[i], K33[i], K34[i], K35[i], K36[i]],
                [K14[i], K24[i], K34[i], K44[i], K45[i], K46[i]],
                [K15[i], K25[i], K35[i], K45[i], K55[i], K56[i]],
                [K16[i], K26[i], K36[i], K46[i], K56[i], K66[i]]
            ])
            K[i,:,:] = k_matrix
            # Assemble inertia matrix at the reference axis
            i_matrix = np.array([
                [mass[i], 0.0, 0.0, 0.0, 0.0, -cm_y[i]*mass[i]],
                [0.0, mass[i], 0.0, 0.0, 0.0, cm_x[i]*mass[i]],
                [0.0, 0.0, mass[i], cm_y[i]*mass[i], -cm_x[i]*mass[i], 0.0],
                [0.0, 0.0, cm_y[i]*mass[i], i_edge[i], -i_cp[i], 0.0],
                [0.0, 0.0, -cm_x[i]*mass[i], -i_cp[i], i_flap[i], 0.0],
                [-cm_y[i]*mass[i], cm_x[i]*mass[i], 0.0, 0.0, 0.0, i_plr[i]]
            ])
            I[i,:,:] = i_matrix

        outputs["K"] = K
        outputs["I"] = I
class KI_to_Elastic(ExplicitComponent):
    def initialize(self):
        self.options.declare('modeling_options')

    def setup(self):
        modopt = self.options['modeling_options']
        rotorse_options  = modopt['WISDEM']['RotorSE']
        self.n_span = n_span = rotorse_options['n_span']

        # Inputs
        self.add_input(
            "K", 
            val=np.zeros((n_span,6,6)), 
            desc="Stiffness matrix at the center of the windIO reference axes."
        )
        self.add_input(
            "I", 
            val=np.zeros((n_span,6,6)), 
            desc="Inertia matrix at the center of the windIO reference axes."
        )
        self.add_input(
            "theta",
            val=np.zeros(n_span),
            units="deg",
            desc="Aerodynamic twist angle at each section (positive decreases angle of attack)",
        )

        # Outputs
        self.add_output("A", val=np.ones(n_span), units="m**2", desc="cross sectional area")
        self.add_output("EA", val=np.zeros(n_span), units="N", desc="Axial stiffness at the elastic center.")
        self.add_output("EIxx", val=np.zeros(n_span), units="N*m**2", desc="Section lag (edgewise) bending stiffness about the XE axis.")
        self.add_output("EIyy", val=np.zeros(n_span), units="N*m**2", desc="Section flap bending stiffness about the YE axis.")
        self.add_output("EIxy", val=np.zeros(n_span), units="N*m**2", desc="Coupled flap-lag stiffness with respect to the XE-YE frame")
        self.add_output("EA_EIxx", val=np.zeros(n_span), units="N*m", desc="Coupled axial-lag stiffness with respect to the XE-YE frame")
        self.add_output("EA_EIyy", val=np.zeros(n_span), units="N*m", desc="Coupled axial-flap stiffness with respect to the XE-YE frame")
        self.add_output("EIxx_GJ", val=np.zeros(n_span), units="N*m**2", desc="Coupled lag-torsion stiffness with respect to the XE-YE frame")
        self.add_output("EIyy_GJ", val=np.zeros(n_span), units="N*m**2", desc="Coupled flap-torsion stiffness with respect to the XE-YE frame ")
        self.add_output("EA_GJ", val=np.zeros(n_span), units="N*m", desc="Coupled axial-torsion stiffness")
        self.add_output("GJ", val=np.zeros(n_span), units="N*m**2", desc="Section torsion stiffness at the elastic center.")

        self.add_output("rhoA", val=np.zeros(n_span), units="kg/m", desc="Section mass per unit length")
        self.add_output("rhoJ", val=np.zeros(n_span), units="kg*m", desc="polar mass moment of inertia per unit length")
        self.add_output(
            "Tw_iner",
            val=np.zeros(n_span),
            units="deg",
            desc="Orientation of the section principal inertia axes with respect the blade reference plane",
        )
        self.add_output(
            "x_tc",
            val=np.zeros(n_span),
            units="m",
            desc="X-coordinate of the tension-center offset with respect to the XR-YR axes",
        )
        self.add_output(
            "y_tc",
            val=np.zeros(n_span),
            units="m",
            desc="Chordwise offset of the section tension-center with respect to the XR-YR axes",
        )
        self.add_output(
            "x_sc",
            val=np.zeros(n_span),
            units="m",
            desc="X-coordinate of the shear-center offset with respect to the XR-YR axes",
        )
        self.add_output(
            "y_sc",
            val=np.zeros(n_span),
            units="m",
            desc="Chordwise offset of the section shear-center with respect to the reference frame, XR-YR",
        )
        self.add_output(
            "x_cg",
            val=np.zeros(n_span),
            units="m",
            desc="X-coordinate of the center-of-mass offset with respect to the XR-YR axes",
        )
        self.add_output(
            "y_cg",
            val=np.zeros(n_span),
            units="m",
            desc="Chordwise offset of the section center of mass with respect to the XR-YR axes",
        )
        self.add_output(
            "flap_iner",
            val=np.zeros(n_span),
            units="kg/m",
            desc="Section flap inertia about the Y_G axis per unit length.",
        )
        self.add_output(
            "edge_iner",
            val=np.zeros(n_span),
            units="kg/m",
            desc="Section lag inertia about the X_G axis per unit length",
        )

        # the following will be given as pseudo 
        self.add_output(
            "xu_spar",
            val=np.zeros(n_span),
            desc="x-position of midpoint of spar cap on upper surface for strain calculation",
        )
        self.add_output(
            "xl_spar",
            val=np.zeros(n_span),
            desc="x-position of midpoint of spar cap on lower surface for strain calculation",
        )
        self.add_output(
            "yu_spar",
            val=np.zeros(n_span),
            desc="y-position of midpoint of spar cap on upper surface for strain calculation",
        )
        self.add_output(
            "yl_spar",
            val=np.zeros(n_span),
            desc="y-position of midpoint of spar cap on lower surface for strain calculation",
        )
        self.add_output(
            "xu_te",
            val=np.zeros(n_span),
            desc="x-position of midpoint of trailing-edge panel on upper surface for strain calculation",
        )
        self.add_output(
            "xl_te",
            val=np.zeros(n_span),
            desc="x-position of midpoint of trailing-edge panel on lower surface for strain calculation",
        )
        self.add_output(
            "yu_te",
            val=np.zeros(n_span),
            desc="y-position of midpoint of trailing-edge panel on upper surface for strain calculation",
        )
        self.add_output(
            "yl_te",
            val=np.zeros(n_span),
            desc="y-position of midpoint of trailing-edge panel on lower surface for strain calculation",
        )
        
    def compute(self, inputs, outputs):

        # Initialize empty 6x6 K and I matrices
        K_tc = np.zeros((self.n_span,6,6))
        I_cg = np.zeros((self.n_span,6,6))
        aero_twist =  np.deg2rad(inputs["theta"])


        for i in range(self.n_span):
            # Get the stiffness matrix at the reference axis
            K_ref = inputs["K"][i,:,:]
            I_ref = inputs["I"][i,:,:]

            # Find shear, elastic/tension center
            # Reference SONATA anbax
            K1 = np.array([[K_ref[n, m] for m in range(3)] for n in range(3)])
            K3 = np.array([[K_ref[n, m+3] for m in range(3)] for n in range(3)])
            try:
                Y = np.linalg.solve(K1, -K3)
                x_sc = -Y[1, 2]
                y_sc = Y[0, 2]

                x_tc = Y[2, 1]
                y_tc = -Y[2, 0]
            except:
                print(f"Failed to compute shear center and tension center at section {i}.")
                if (K_ref[3,3] == 0) and (K_ref[4,4] == 0):
                    raise Exception("K44 and K55 cannot be zeros!")
                else:
                    x_sc = 0
                    y_sc = 0

                    x_tc = 0
                    y_tc = 0

                    print("Check if K11, K22, K33 are non-zeros and correct. Shear center and tension center valuse are set to zeros.") 
            # Transform stiffness matrix to elastic center/tension center
            transform = TransformCrossSectionMatrix()
            T = transform.CrossSectionTranslationMatrix(x_tc, y_tc)
            K_tc = T.T @ K_ref @ T

            # assign stiffness matrix to outputs
            outputs["EA"][i] = K_tc[2,2]
            outputs["EA_EIxx"][i] = K_tc[2,3]
            outputs["EA_EIyy"][i] = K_tc[2,4]
            outputs["EIxx"][i] = K_tc[3,3]
            outputs["EIyy"][i] = K_tc[4,4]
            outputs["EIxy"][i] = K_tc[3,4]
            outputs["GJ"][i] = K_tc[5,5]
            outputs["EA_GJ"][i] = K_tc[2,5]
            outputs["EIxx_GJ"][i] = K_tc[3,5]
            outputs["EIyy_GJ"][i] = K_tc[4,5]
            outputs["x_tc"][i] = x_tc
            outputs["y_tc"][i] = y_tc
            outputs["x_sc"][i] = x_sc
            outputs["y_sc"][i] = y_sc

            # Transform mass matrix to inertia frame
            x_cg = I_ref[1,5]/I_ref[0,0]
            y_cg = -I_ref[0,5]/I_ref[0,0]


            transform = TransformCrossSectionMatrix()
            T = transform.CrossSectionTranslationMatrix(x_cg, y_cg)
            I_cg = T.T @ I_ref @ T
            # find the inertia twist
            I3 = np.array([[I_cg[n+3, m+3] for m in range(3)] for n in range(3)]) 
            (w3, v3) = np.linalg.eig(I3)

            # This angle solve is likely working only within in [-pi/2, pi/2]
            if np.abs(v3[0,0]) < np.abs(v3[0,1]):
                angle = np.arccos(v3[0,0])
            else:
                angle = -np.arcsin(v3[0,1])
            R = transform.CrossSectionRotationMatrix(-angle) # negative because from ref to inertia
            I_cg = R.T @ I_cg @ R

            # assign inertia matrix to outputs
            
            outputs["rhoA"][i] = I_cg[0,0]
            outputs["rhoJ"][i] = I_cg[5,5] # This is actually J - edge+flap
            outputs["edge_iner"][i] = I_cg[3,3]
            outputs["flap_iner"][i] = I_cg[4,4]
            outputs["Tw_iner"][i] = np.rad2deg(aero_twist[i]-angle)
            outputs["x_cg"][i] = x_cg
            outputs["y_cg"][i] = y_cg

            # Aproximate area A
            E_est_xx = outputs["EIxx"][i] / outputs["edge_iner"][i]
            E_est_yy = outputs["EIyy"][i] / outputs["flap_iner"][i]
            A_est_xx = outputs["EA"][i] / E_est_xx
            A_est_yy = outputs["EA"][i] / E_est_yy
            outputs["A"][i] = 0.5*(A_est_xx+A_est_yy) # Approximate the area          


class RotorElasticity(Group):
    # OpenMDAO group to compute the blade elastic properties and natural frequencies
    def initialize(self):
        self.options.declare("modeling_options")
        self.options.declare("opt_options")

    def setup(self):
        modeling_options = self.options["modeling_options"]
        opt_options = self.options["opt_options"]

        # Get elastic properties by running precomp
        promote_list = [
            "EA",
            "EIxx",
            "EIyy",
            "EIxy",
            "GJ",       
            "EA_EIxx",
            "EA_EIyy",
            "EIxx_GJ",
            "EIyy_GJ",
            "EA_GJ",
            "rhoA",
            "rhoJ",
            "Tw_iner",
            "x_tc",
            "y_tc",
            "x_cg",
            "y_cg",
            "edge_iner",
            "flap_iner",
        ]

        if modeling_options["user_elastic"]["blade"]:
            self.add_subsystem("generate_KI", assemble_KI(modeling_options=modeling_options), promotes=['*'])
            self.add_subsystem("precomp", KI_to_Elastic(modeling_options=modeling_options), promotes=['*'])

        else:

            self.add_subsystem(
                "precomp",
                RunPreComp(modeling_options=modeling_options, opt_options=opt_options),
                promotes=promote_list
                + [
                    "r",
                    "chord",
                    "theta",
                    "A",
                    "precurve",
                    "presweep",
                    "pitch_axis",
                    "coord_xy_interp",
                    "sc_ss_mats",
                    "sc_ps_mats",
                    "te_ss_mats",
                    "te_ps_mats",
                    "xu_spar",
                    "xl_spar",
                    "yu_spar",
                    "yl_spar",
                    "xu_te",
                    "xl_te",
                    "yu_te",
                    "yl_te",
                    "n_blades",
                ],
            )

            self.add_subsystem("generate_KI", generate_KI(modeling_options=modeling_options), promotes=promote_list+["K", "I"])

        # Compute total blade properties
        self.add_subsystem("total_blade_properties",
                           TotalBladeProperties(modeling_options=modeling_options, opt_options=opt_options),
                           promotes=["*"])
            





from scipy.special import gamma

import numpy as np
import openmdao.api as om

# ---------------------
# Map Design Variables to Discretization
# ---------------------


class CDFBase(om.ExplicitComponent):
    """cumulative distribution function"""

    def initialize(self):
        self.options.declare("nspline")

    def setup(self):
        nspline = self.options["nspline"]

        self.add_input("x", shape=nspline, units="m/s", desc="corresponding reference height")
        self.add_input("k", shape=1, desc="shape or form factor")

        self.add_output("F", shape=nspline, units="m/s", desc="magnitude of wind speed at each z location")


class WeibullCDF(CDFBase):
    def setup(self):
        super(WeibullCDF, self).setup()
        """Weibull cumulative distribution function"""

        self.add_input("A", shape=1, desc="scale factor")

        self.declare_partials("F", "x")
        self.declare_partials("F", ["A", "k"])

    def compute(self, inputs, outputs):
        outputs["F"] = 1.0 - np.exp(-((inputs["x"] / inputs["A"]) ** inputs["k"]))

    def compute_partials(self, inputs, J):
        x = inputs["x"]
        A = inputs["A"]
        k = inputs["k"]

        J["F", "x"] = np.diag(np.exp(-((x / A) ** k)) * (x / A) ** (k - 1) * k / A)
        J["F", "A"] = -k / A * np.exp(-((x / A) ** k)) * (x / A) ** k
        J["F", "k"] = np.exp(-((x / A) ** k)) * (x / A) ** k * np.log(x / A)


class WeibullWithMeanCDF(CDFBase):
    def setup(self):
        super(WeibullWithMeanCDF, self).setup()
        """Weibull cumulative distribution function"""

        self.add_input("xbar", shape=1, units="m/s", desc="mean value of distribution")

        self.declare_partials("F", "x")
        self.declare_partials("F", "xbar")
        self.declare_partials("F", "k", method="fd")

    def compute(self, inputs, outputs):
        A = inputs["xbar"] / gamma(1.0 + 1.0 / inputs["k"])

        outputs["F"] = 1.0 - np.exp(-((inputs["x"] / A) ** inputs["k"]))

    def compute_partials(self, inputs, J):
        x = inputs["x"]
        k = inputs["k"]
        A = inputs["xbar"] / gamma(1.0 + 1.0 / k)
        dx = np.diag(np.exp(-((x / A) ** k)) * (x / A) ** (k - 1) * k / A)
        dxbar = -np.exp(-((x / A) ** k)) * (x / A) ** (k - 1) * k * x / A**2 / gamma(1.0 + 1.0 / k)

        J["F", "x"] = dx
        J["F", "xbar"] = dxbar


def RayleighCDF_func(x, xbar=10.0):
    return 1.0 - np.exp(-np.pi / 4.0 * (x / xbar) ** 2)


class RayleighCDF(CDFBase):
    def setup(self):
        super(RayleighCDF, self).setup()

        # variables
        self.add_input("xbar", shape=1, units="m/s", desc="reference wind speed (usually at hub height)")

        self.declare_partials("F", "x")
        self.declare_partials("F", "xbar")

    def compute(self, inputs, outputs):
        outputs["F"] = RayleighCDF_func(inputs["x"], inputs["xbar"])

    def compute_partials(self, inputs, J):
        x = inputs["x"]
        xbar = inputs["xbar"]
        dx = np.diag(np.exp(-np.pi / 4.0 * (x / xbar) ** 2) * np.pi * x / (2.0 * xbar**2))
        dxbar = -np.exp(-np.pi / 4.0 * (x / xbar) ** 2) * np.pi * x**2 / (2.0 * xbar**3)

        J["F", "x"] = dx
        J["F", "xbar"] = dxbar
